/**********************************************************************
 * LeechCraft - modular cross-platform feature rich internet client.
 * Copyright (C) 2006-2014  Georg Rudoy
 *
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************/

#pragma once

#include <functional>
#include <atomic>
#include <QThread>
#include <QMutex>
#include <QMutexLocker>
#include <QFutureInterface>
#include <QFuture>
#include <QList>
#include "futures.h"
#include "threadsconfig.h"

namespace LeechCraft
{
namespace Util
{
	class UTIL_THREADS_API WorkerThreadBase : public QThread
	{
		Q_OBJECT

		std::atomic_bool IsPaused_ { false };

		QMutex FunctionsMutex_;
		QList<std::function<void ()>> Functions_;
	public:
		using QThread::QThread;

		void SetPaused (bool);

		template<typename F>
		QFuture<ResultOf_t<F ()>> ScheduleImpl (const F& func)
		{
			QFutureInterface<ResultOf_t<F ()>> iface;
			iface.reportStarted ();

			auto reporting = [func, iface] () mutable
			{
				ReportFutureResult (iface, func);
			};

			{
				QMutexLocker locker { &FunctionsMutex_ };
				Functions_ << reporting;
			}

			emit rotateFuncs ();

			return iface.future ();
		}

		template<typename F, typename... Args>
		QFuture<ResultOf_t<F (Args...)>> ScheduleImpl (const F& f, Args&&... args)
		{
			return ScheduleImpl ([f, args...] { return Invoke (f, args...); });
		}
	protected:
		void run () override final;

		virtual void Initialize () = 0;
		virtual void Cleanup () = 0;
	private:
		void RotateFuncs ();
	signals:
		void rotateFuncs ();
	};

	template<typename WorkerType>
	class WorkerThread : public WorkerThreadBase
	{
	protected:
		std::shared_ptr<WorkerType> Worker_;
	public:
		using WorkerThreadBase::WorkerThreadBase;

		using WorkerThreadBase::ScheduleImpl;

		template<typename F, typename... Args>
		QFuture<ResultOf_t<F (WorkerType*, Args...)>> ScheduleImpl (const F& f, Args&&... args)
		{
			const auto fWrapped = [f, this] (auto... args) { return Invoke (f, Worker_.get (), args...); };
			return WorkerThreadBase::ScheduleImpl (fWrapped, std::forward<Args> (args)...);
		}
	protected:
		void Initialize () override
		{
			Worker_ = std::make_shared<WorkerType> ();
		}

		void Cleanup () override
		{
			Worker_.reset ();
		}
	};
}
}
